# -*- coding: utf-8 -*-
"""

@author: eendebakpt
"""

#%% Load packages
import numpy as np
import oalib
import oapackage
import sys


#%%


def oa2graph(al, adata, verbose=1):
    """
    Convert orthogonal array to graph representation

    The conversion method is as in Ryan and Bulutoglu.
    The resulting graph is bi-partite.
    The graph representation can be used for isomorphism testing.

    Example
    -------
    >     im, colors, r = oa2graph( A, oadata );

    """
    #A = np.array(al)
    A = al.getarray(verbose=0)
    #print('type A: %s' % type(A) )
    nrows = adata.N
    ncols = A.shape[1]
    nRowVertices = A.shape[0]
    nColumnLevelVertices = sum(adata.getS())
    nVertices = adata.N + ncols + nColumnLevelVertices
    nColVertices = ncols
    colOffset = adata.N

    s = np.array(adata.getS())
    sc = np.cumsum(s)
    sc0 = np.hstack(([0], sc))
    qq = np.ones(nColumnLevelVertices)  # colors for column level vertices
    for ii in range(sc.size):
        qq[sc0[ii]:sc0[ii + 1]] = 2 + ii
    qq = 2 * np.ones(nColumnLevelVertices)  # colors for column level vertices

    vertexOffsets = adata.N + ncols + np.hstack((0, s[0:-1])).cumsum()
    colors = np.hstack(
        (np.zeros(adata.N), np.ones(ncols), qq))

    im = np.zeros((nVertices, nVertices))  # incidence matrix

    for row in range(0, nrows):
        idx = A[row, :] + vertexOffsets
        im[row, idx] = 1
        im[idx, row] = 1

    if nColVertices > 0:
        colidx = 2
        for col in range(0, ncols):
            sel = vertexOffsets[col] + range(0, s[col])
            im[colOffset + col, sel] = colidx
            im[sel, colOffset + col] = colidx

    # The non-row vertices do not have any connections to other non-row
    # vertices.

    if 0:
        # xy is an old data structure which is not used any moref
        xy = np.zeros((2, nrows + s.sum()))

        # calculate positions
        for row in range(0, nrows):
            xy[:, row] = np.array([0, row])

        pos = nrows
        for col in range(0, ncols):
            for ss in range(0, s[col]):
                xy[:, pos] = np.array([2 + ss / s[col], col])
                pos = pos + 1

    return im, colors, dict({'adata': adata, 'im': im, 'colors': colors, 'nVertices': nVertices})


import oapackage


def graph2arrayTransformation(pp, arrayclass, verbose=0):
    """ From a relabelling of the graph return the corresponding array transformation """
    ppi = np.zeros(len(pp), )
    ppi[pp] = range(len(pp))
    ppi = np.array(ppi).astype(int)
    pp = np.array(pp).astype(int)

    # extract colperms and rowperm and levelperms from this...

    rowperm = np.array((pp[0:arrayclass.N]))
    rowperm = rowperm - rowperm.min()
    colperm = np.array((pp[arrayclass.N:(arrayclass.N + arrayclass.ncols)]))
    colperm = np.argsort(colperm)  # colperm-colperm.min()
    ns = np.sum(arrayclass.getS())
    lvlperm = np.array(
        (pp[(arrayclass.N + arrayclass.ncols):(arrayclass.N + arrayclass.ncols + ns)]))
    lvlperm = lvlperm - lvlperm.min()

    #%%
    ttr = oapackage.array_transformation_t(arrayclass)
    ttr.setrowperm(rowperm)
    ttr = ttr.inverse()
    ttc = oapackage.array_transformation_t(arrayclass)
    ttc.setcolperm(colperm)

    ttl = oapackage.array_transformation_t(arrayclass)

    ncols = arrayclass.ncols
    cs = np.hstack(([0], np.cumsum(arrayclass.getS())))
    lp = []
    for ii in range(ncols):
        ww = lvlperm[cs[ii]:cs[ii + 1]]
        ww = ww - ww.min()
        ww = np.argsort(ww)
        lp.append(ww)
        ttl.setlevelperm(ii, ww)

    ttl = ttl.inverse()

    tt = ttr * ttc * ttl
    return tt

#


#%%

from oapackage import makearraylink


def selectIsomorphismClasses(sols, verbose=1):
    """ Select isomorphism classes from a list of designs 

    Args:
        sols (list of arrays)
        verbose (int)
    Return:
        indices (list of integers): indices of the isomorphism classes
        mm (list of arrays): the arrays in normal form


    Example:
        >>> sols=[oapackage.exampleArray(i) for i in [26,26,27,26]]
        >>> idx, mm = selectIsomorphismClasses(sols)

    To select one representative array from each isomorphism class one can use:
        >>> _, ui = np.unique(idx, return_index=True)
        >>> representatives = [sols[i] for i in ui]

    """

    # perform check on array data type
    mm = []
    for ii, al in enumerate(sols):
        if verbose:
            oapackage.tprint('selectIsomorphismClasses: process aray %d/%d' % (ii, len(sols)))
        al = oapackage.makearraylink(al)

        tt = oapackage.reduceOAnauty(al, verbose >= 2)

        #pp, tt = reduceBliss(al, arrayclass, verbose >= 2)
        #tt = graph2arrayTransformation(pp, arrayclass)
        alx = tt.apply(al)
        mm.append(np.array(alx))
        pass

    # perform uniqueness check
    nn = len(mm)
    qq = np.array([None] * nn, dtype=object)
    for ii in range(nn):
        qq[ii] = mm[ii].flatten()

    # Trick to make unique work...
    nx = qq[0].size
    dt = qq[0].dtype.descr * nx
    qqq = np.array(qq, dtype=dt)

    a, indices = np.unique(qqq, return_inverse=True)

    if verbose >= 1:
        print('selectIsomorphismClasses: reduce %d to %d' %
              (len(sols), np.unique(indices).size))

    return indices, mm


def test_select_isomorphism():
    ll = [oapackage.exampleArray(0), oapackage.exampleArray(0)]
    selectIsomorphismClasses(ll)



# -*- coding: utf-8 -*-
"""

Collection of helper functions for OA package

Pieter Eendebak <pieter.eendebak@gmail.com>

@author: eendebakpt
"""

from __future__ import print_function

import oalib
import os
import numpy as np
import time
try:
    import matplotlib
    import matplotlib.pyplot as plt
except:
    # print('Doptim: matplotlib cannot be found, not all functionality is
    # available')
    pass
import oapackage.markup as markup
import oapackage.oahelper as oahelper

from oapackage.markup import oneliner as e

#%%


def array2Dtable(sols, verbose=1, titlestr=None):
    """ Generate HTML table with information about for a list of designs """
    # page.p()
    na = len(sols)
    page = markup.page()
    page.table(style=' border-collapse: collapse;')
    page.tr(style='font-weight: bold; border-bottom: solid 1px black;')
    page.th('Array', style='padding-right:30px; ')
    page.th(('D-efficiency', 'Ds-efficiency', 'D1-efficiency'),
            style='padding-right:14px;')
    page.th(('GWLP'), style='padding-right:14px;')
    page.tr.close()
    for ii, al in enumerate(sols):
        aidx = ii
        (D, Ds, D1) = al.Defficiencies()
        gwlp = al.GWLP()
        page.tr(style='font-weight: normal;')
        page.td('%d' % aidx, style='padding-right:10px;')
        for v in [D, Ds, D1]:
            page.td('%.4f' % v, style='padding-right:1px;')
        gstr = oahelper.gwlp2str(gwlp)
        page.td(e.small(gstr), style='padding-right:1px;')
        page.tr.close()
    page.table.close()
    # page.p.close()
    return page

#%%

try:
    import brewer2mpl
except:
    pass


def generateDscatter(dds, si=0, fi=1, lbls=None, ndata=3, nofig=False, fig=20, scatterarea=80):
    """ Generate scatter plot for D and Ds efficiencies """
    data = dds.T
    pp = oahelper.createPareto(dds)
    paretoidx = np.array(pp.allindices())

    nn = dds.shape[0]
    area = scatterarea * np.ones(nn,) / 2
    area[np.array(pp.allindices())] = scatterarea
    alpha = 1.0

    if dds.shape[1] > ndata:
        colors = dds[:, ndata]
    else:
        colors = np.zeros((nn, 1))

    idx = np.unique(colors).astype(int)

    try:
        mycmap = brewer2mpl.get_map('Set1', 'qualitative', idx.size).mpl_colors
    except:
        mycmap = [matplotlib.cm.jet(ii) for ii in range(4)]
        pass

    # For remaining spines, thin out their line and change the black to a
    # slightly off-black dark grey
    almost_black = '#202020'

    figh = plt.figure(fig)  # ,facecolor='red')
    plt.clf()
    figh.set_facecolor('w')
    ax = plt.subplot(111)

    nonparetoidx = np.setdiff1d(range(nn), paretoidx)

    ax.scatter(data[fi, nonparetoidx], data[si, nonparetoidx], s=.33 * scatterarea,
               c=(.5, .5, .5), linewidths=0, alpha=alpha, label='Non-pareto design')

    # print(area)
    for jj, ii in enumerate(idx):
        gidx = (colors == ii).nonzero()[0]
        gp = np.intersect1d(paretoidx, gidx)

        color = mycmap[jj]
        cc = [color] * len(gp)
        print('index %d: %d points' % (ii, gidx.size))
        ax.scatter(data[fi, gp], data[si, gp], s=scatterarea, c=cc,
                   linewidths=0, alpha=alpha, label=lbls[jj])  # , zorder=4)
        plt.draw()

    if data[si, :].std() < 1e-3:
        y_formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)
        ax.yaxis.set_major_formatter(y_formatter)

    if 0:
        for xi, al in enumerate(sols):
            D, Ds, D1 = al.Defficiencies()
            print('D1 %f Ds %f D %f' % (D1, Ds, D))

            tmp = plt.scatter(Ds, D, s=60, color='r')
            if xi == 0:
                tmp = plt.scatter(Ds, D, s=60, color='r', label='Strength 3')
        plt.draw()

    xlabelhandle = plt.xlabel('$D_s$-efficiency', fontsize=16)
    plt.ylabel('D-efficiency', fontsize=16)

    try:
        oahelper.setWindowRectangle(10, 10, 860, 600)
    except Exception as e:
        print('generateDscatter: setWindowRectangle failed')
        # print(e)
        pass

    plt.axis('image')
    pltlegend = ax.legend(loc=3, scatterpoints=1)  # , fontcolor=almost_black)
    if not nofig:
        plt.show()
    # time.sleep(0.01)
    ax.grid(b=True, which='both', color='0.85', linestyle='-')
    ax.set_axisbelow(True)

    if not nofig:
        plt.draw()

    hh = dict({'ax': ax, 'xlabelhandle': xlabelhandle, 'pltlegend': pltlegend})
    return hh
#%%


def generateDpage(outputdir, arrayclass, dds, allarrays, fig=20, optimfunc=[1, 0, 0], nofig=False, urlprefix='', makeheader=True, verbose=1, lbls=None):

    #%% Prepare data
    if verbose:
        print('generateDpage: dds %s' % str(dds.shape))

    pp = oahelper.createPareto(dds)
    #paretoidx = np.array(pp.allindices())

    narrays = dds.shape[0]
    npareto = pp.number()

    if verbose:
        print('generateDpage: narrays %d' % narrays)

    xstr = oahelper.series2htmlstr(arrayclass, case=1)
    xstrplain = oahelper.series2htmlstr(arrayclass, html=0, case=1)

    if verbose:
        print('generateDpage: selectParetoArrays ')
    paretoarrays = oahelper.selectParetoArrays(allarrays, pp)
    at = array2Dtable(paretoarrays, verbose=1)

    if verbose:
        print('generateDpage: write file with Pareto arrays')

    pfile0 = 'paretoarrays.oa'
    pfile = os.path.join(outputdir, pfile0)
    oalib.writearrayfile(pfile, paretoarrays)

    istrlnk = markup.oneliner.a('paretoarrays.oa', href=urlprefix + pfile0)

    # lbls= ['Optimization of $D+.5 D_s$', 'Optimization of $D+  0.5 D_s$',
    # 'Optimization of $D+3*Ds$', 'Optimization of $D+3*D_s$']
    if lbls is None:
        lbls = ['Optimization of $D$']
    hh = generateDscatter(dds, lbls=lbls, fig=fig, nofig=nofig)
    oahelper.niceplot(hh.get('ax', None), despine=True, legend=hh['pltlegend'])

    scatterfile = os.path.join(outputdir, 'scatterplot.png')
    if verbose:
        print('generateDpage: writen scatterplot to %s' % scatterfile)
    plt.savefig(scatterfile, bbox_inches='tight', pad_inches=0.25, dpi=160)

    #%% Create page

    page = markup.page()

    if makeheader:
        page.init(title="Class %s" % xstrplain,
                  css=('../oastyle.css'),
                  lang='en', htmlattrs=dict({'xmlns': 'http://www.w3.org/1999/xhtml', 'xml:lang': 'en'}),
                  header="<!-- Start of page -->",
                  bodyattrs=dict({'style': 'padding-left: 3px;'}),
                  # doctype=markup.doctype.strict,
                       doctype='<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',
                       metainfo=({'text/html': 'charset=utf-8', 'keywords': 'orthogonal arrays designs',
                                  'robots': 'index, follow', 'description': 'Even-Odd arrays'}),
                       footer="<!-- End of page -->")

    page.h1('Results for array class %s ' % xstr)

    # mathjax is not updated properly...
    ss = 'The Pareto optimaly was calculated according to the statistics \(D\), \(D1\) and \(Ds\).'
    ss = 'The Pareto optimaly was calculated according to the statistics D, D<sub>1</sub> and D<sub>s</sub>.'
    if npareto == 1:
        page.p('Generated %d arrays, %d is Pareto optimal. %s' %
               (narrays, npareto, ss))
    else:
        page.p('Generated %d arrays, %d are Pareto optimal. %s' %
               (narrays, npareto, ss))

    if narrays > 0:

        scores = calcScore(dds, optimfunc)
        tmp, dd, sols = selectDn(scores, dds, allarrays, nout=1)
        A = sols[0]
        bestdesignfile = os.path.join(outputdir, 'best-design.oa')
        oalib.writearrayfile(bestdesignfile, A)

        page.h2('Best design')
        page.p('The best design: %s.' %
               e.a('best-design.oa', href=os.path.join(urlprefix, 'best-design.oa')))

        page.p()

        dd = dd[0]
        page.span('D-efficiency: %.4f, ' % A.Defficiency())  # page.br()
        page.span('D<sub>s</sub>-efficiency: %.4f, ' % dd[1])  # page.br()
        page.span('D<sub>1</sub>-efficiency: %.4f' % dd[2])
        page.br()
        page.span('A-efficiency: %.3f' % A.Aefficiency())
        page.br()
        gwlp = A.GWLP()
        # gwlp=','.join(['%.3f' % xq for xq in gwlp])
        gwlp = oahelper.gwlp2str(gwlp, jstr=', ')
        page.span('Generalized wordlength pattern: %s' % gwlp)
        page.br()
        # page.p('D-efficiency: %.3f' % A.Defficiency() )
        pec = oalib.PECsequence(A)
        pec = ','.join(['%.3f' % xq for xq in pec])
        page.span('PEC-sequence: %s' % pec)
        page.br()

    page.h2('Table of Pareto optimal arrays ')

    page.span(str(at))
    page.p('All Pareto optimal arrays: %s' % istrlnk)

    page.img(src=urlprefix + 'scatterplot.png',
             style="margin: 10px; width:95%; min-width: 300px;  max-width:1100px; height: auto; ")

    citationstr = markup.oneliner.a(
        'Complete Enumeration of Pure-Level and Mixed-Level Orthogonal Arrays', href='http://dx.doi.org/10.1002/jcd.20236')

    page.br(clear='both')
    page.p(
        'Citation notice: if you make use of the results on this page, please cite the following paper:')
    page.p('%s, Journal of Combinatorial Designs, Volume 18, Issue 2, pages 123-140, 2010.' %
           citationstr)

    page.p('Generated with oapackage %s, date %s.' %
           (oalib.version(), oahelper.timeString()))

    outfile = os.path.join(outputdir, 'Dresults.html')
    fid = open(outfile, 'wt')
    fid.write(str(page))
    fid.close()
    print('written to file %s' % outfile)

    return outfile

# generateDpage(doptimdir, arrayclass, dds, sols, optimfunc=[1,0,0],
# makeheader=True)

#%%

# TODO: implement fast update of xf


def optimDeffhelper(classdata):
    """ Helper function that is suitable for the multi-processing framework """

    N = classdata[0]
    k = classdata[1]
    alpha = classdata[2]
    method = classdata[3]
    p = classdata[4]
    nabort = p.get('nabort', 2500)
    niter = p.get('nabort', 12000)

    arrayclass = oalib.arraydata_t(2, N, 1, k)
    al = arrayclass.randomarray(1)

    vv = optimDeffPython(
        al, niter=niter, nabort=nabort, verbose=0, alpha=alpha, method=method)
    return vv[0], vv[1].getarray()


def calcScore(dds, optimfunc):
    """ Calculate D-efficiency score using multiple efficiencies and a weight factor """
    n = dds.shape[0]
    scores = np.zeros(n, )
    for ii, d in enumerate(dds):
        alpha = optimfunc
        scores[ii] = alpha[0] * d[0] + alpha[1] * d[1] + alpha[2] * d[2]
    return scores


def optimDeffPython(A0, arrayclass=None, niter=10000, nabort=2500, verbose=1, alpha=[1, 0, 0], method=0):
    """ Optimize arrays """
    # get factor levels
    if arrayclass is None:
        s = A0.getarray().max(axis=0) + 1
    else:
        s = arrayclass.getS()
        arrayclass
    sx = tuple(s.astype(np.int64))
    sx = tuple([int(v) for v in sx])

    sg = oalib.symmetry_group(sx)
    gidx = tuple(sg.gidx)
    gsize = tuple(sg.gsize)
    gstart = tuple(sg.gstart)

    nx = 0

    Dinitial = A0.Defficiency()
    if verbose:
        print('optimDeff: initial Deff %.4f' % Dinitial)
    N = A0.n_rows
    k = A0.n_columns

    # initialize score
    if str(type(alpha)) == "<type 'function'>":
        d = alpha(A0)
    else:
        D, Ds, D1 = A0.Defficiencies()
        d = alpha[0] * D + alpha[1] * Ds + alpha[2] * D1
    A = A0.clone()
    lc = 0
    for ii in range(0, niter):
        # r=random.randint(0,N-1)
        # c=random.randint(0,k-1)
        r = np.random.randint(N)
        c = np.random.randint(k)
        r2 = np.random.randint(N)
        # make sure c2 is in proper column
        # c2=np.random.randint(k)
        c2 = gstart[gidx[c]] + oalib.fastrand() % gsize[gidx[c]]
        # print('  %d,%d <--> %d,%d' % (r,c,r2,c2))

        # o=A[r, c]; o2=A[r2, c2]
        o = A._at(r, c)
        o2 = A._at(r2, c2)  # no extra error checking
        # swap
        if method == oalib.DOPTIM_SWAP:
            # swap
            if o == o2:
                continue
            A._setvalue(r, c, o2)
            A._setvalue(r2, c2, o)
        elif method == oalib.DOPTIM_NONE:
            tmp = 0
        else:
            # flip
            A._setvalue(r, c, 1 - o)

        if str(type(alpha)) == "<type 'function'>":
            dn = alpha(A)
        else:
            D, Ds, D1 = A.Defficiencies()
            nx = nx + 1
            # print(alpha)
            dn = alpha[0] * D + alpha[1] * Ds + alpha[2] * D1
        if (dn >= d):
            if dn > d:
                lc = ii
                d = dn
            if verbose >= 2:
                print('ii %d: %.6f -> %.6f' % (ii, d, dn))

        else:
            # restore to original
            if method == oalib.DOPTIM_SWAP:
                A._setvalue(r, c, o)
                A._setvalue(r2, c2, o2)
            elif method == oalib.DOPTIM_NONE:
                tmp = 0
            else:
                A._setvalue(r, c, o)
        if (ii - lc) > nabort:
            if verbose:
                print('optimDeff: early abort ii %d, lc %d' % (ii, lc))
            break

    Dfinal = A.Defficiency()

    if verbose:
        if Dfinal > Dinitial:
            print('optimDeff: final Deff improved: %.4f -> %.4f' %
                  (Dinitial, A.Defficiency()))
        else:
            print('optimDeff: final Deff %.4f' % A.Defficiency())

    if verbose >= 2:
        print('nx %d' % nx)
    return d, A


#%%
def filterPareto(scores, dds, sols, verbose=0):
    """ From a list of designs select only the pareto optimal designs """
    pp = oahelper.createPareto(dds, verbose=0)
    paretoidx = np.array(pp.allindices())

    pscores = scores[paretoidx]
    pdds = dds[paretoidx]
    psols = [sols[i] for i in paretoidx]

    return pscores, pdds, psols

#%%


def scoreDn(dds, optimfunc):
    """ Calculate scores from various efficiencies """
    scores = np.array([oalib.scoreD(dd, optimfunc) for dd in dds])
    return scores


def selectDn(scores, dds, sols, nout=1, sortfull=True):
    """ Select best arrays according to given scores
        The resulting data is sorted

    Arguments
    ---------
    scores : array
        scores for the designs
    dds : array
        array with efficiencies of the designs
    sols : list
        list of designs
    nout : integer or None
        Number of results to return. None means return all results

    Returns
    -------
    scores, dds, sols : sorted arrays
    """

    if len(sols) == 0:
        return scores, dds, sols

    if sortfull:
        idx = np.lexsort([dds[:, ii]
                          for ii in range(dds.shape[1])[::-1]] + [scores])[::-1]
    else:
        idx = np.argsort(-scores.flatten())
    # print(idx.shape)
    # print(scores.shape)
    scores = scores[idx]
    dds = dds[idx, :]
    sols = [sols[ii] for ii in idx]
    if not nout is None:
        # sort the arrays
        nout = np.minimum(nout, scores.size)
        scores = scores[0:nout]
        dds = dds[range(nout), :]
        sols = sols[0:nout]
    return scores, dds, sols


def Doptimize(arrayclass, nrestarts=10, optimfunc=[1, 0, 0], verbose=1, maxtime=180, selectpareto=True, nout=None, method=oalib.DOPTIM_UPDATE, niter=100000, nabort=0, dverbose=1):
    """ Calculate D-optimal designs


    For more details see the paper "Two-Level Designs to Estimate All Main
    Effects and Two-Factor Interactions", http://dx.doi.org/10.1080/00401706.2016.1142903

    Arguments
    ---------
    arrayclass : object
        Specifies the type of design to optimize
    nrestarts : integer
        Number of restarts of the algorithm
    optimfunc : list with 3 floats
        Gives the optimization weights
    verbose : integer
        A higher numer gives more output
    maxtime: float
        Maximum running time of the algorithm
    selectpareto : boolean, default is True
        If True then only the Pareto optimal designs are returned
    nout : integer, default None
        Number of designs to return. If None,  return all designs

    Returns
    -------
    scores, dds, designs, nrestarts


    """
    if verbose:
        print('Doptim: optimization class %s' % arrayclass.idstr())
    t0 = time.time()

    if optimfunc is None:
        optimfunc = [1, 2, 0]

    if 1 and isinstance(optimfunc, list):
        rr = oalib.Doptimize(arrayclass, nrestarts, alpha=optimfunc, verbose=dverbose,
                             method=method, niter=niter, maxtime=maxtime, nabort=nabort)
        dds, sols = rr.dds, rr.designs
        dds = np.array([x for x in dds])
        # needed because of SWIG wrapping of struct type
        sols = [x.clone() for x in sols]
        nrestarts = rr.nrestarts
        # dds,sols=rr[0],rr[1]
        # dds=rr[0]
        # sols=[x.clone() for x in sols]
        # dt=time.time()-t0
        scores = np.array(
            [oalib.scoreD(A.Defficiencies(), optimfunc) for A in sols])

        if verbose >= 3:
            print('Doptimize: max score %.3f, max D: %.6f' %
                  (np.max(scores), np.max([A.Defficiency() for A in sols])))
        # dt=time.time()-t0

    else:
        raise Exception('code not tested....')
        scores = np.zeros((0, 1))
        dds = np.zeros((0, 3))
        sols = []  # oalib.arraylist_t()

        nrestarts = 0
        for ii in range(nrestarts):
            if verbose:
                oahelper.tprint('Doptim: iteration %d/%d (time %.1f/%.1f)' %
                                (ii, nrestarts, time.time() - t0, maxtime), dt=4)
            al = arrayclass.randomarray(1)

            if isinstance(optimfunc, list):
                alpha = optimfunc
                Ax = oalib.optimDeff(
                    al, arrayclass, alpha, verbose >= 2, method, niter, nabort)
                dd = Ax.Defficiencies()
                score = oalib.scoreD(dd, optimfunc)
            else:
                score, Ax = optimDeffPython(
                    al, verbose=0, niter=niter, alpha=optimfunc, method=method, nabort=nabort)
                dd = Ax.Defficiencies()

            if time.time() - t0 > maxtime:
                if verbose:
                    print('Doptim: max time exceeded, aborting')
                break

            scores = np.vstack((scores, [score]))
            dds = np.vstack((dds, dd))
            sols.append(Ax)
            nrestarts = nrestarts + 1

            if verbose >= 2:
                print('  generated array: %f %f %f' % (dd[0], dd[1], dd[2]))

            if selectpareto and ii % 502 == 0:
                scores, dds, sols = filterPareto(scores, dds, sols)

            # print(dds.shape)

    if selectpareto:
        if verbose >= 2:
            print('Doptim: before Pareto filter (%d arrays)' % len(sols))
        scores, dds, sols = filterPareto(scores, dds, sols)

    if verbose:
        dt = time.time() - t0
        print('Doptim: done (%d arrays, %.1f [s])' % (len(sols), dt))
        # print(sols)

    # sort & select
    scores, dds, sols = selectDn(scores, dds, sols, nout=nout)

    return scores, dds, sols, nrestarts



# -*- coding: utf-8 -*-
"""

Collection of helper functions for OA package

Pieter Eendebak <pieter.eendebak@gmail.com>

@author: eendebakpt
"""

#%% Load packages
from __future__ import print_function

import oapackage
import oalib
import sys
import os
import numpy as np
import functools
from collections import Counter
import operator
import fileinput
import traceback
import re
from time import gmtime, strftime
import time
import warnings


try:
    import matplotlib
    import matplotlib.pyplot as plt
except:
    warnings.warn(
        'oahelper: matplotlib cannot be found, not all functionality is available')
    pass


#%% Load Qt support

try:
    try:
        from PySide import QtGui
    except:
        from PyQt4 import QtGui

    _applocalqt = None

    def monitorSizes(verbose=0):
        """ Return monitor sizes """
        if sys.platform == 'win32':
            import ctypes
            user32 = ctypes.windll.user32
            wa = [
                [0, 0, user32.GetSystemMetrics(0), user32.GetSystemMetrics(1)]]
        else:
            _applocalqt = QtGui.QApplication.instance()
            if _applocalqt is None:
                _applocalqt = QtGui.QApplication([])
                _qd = QtGui.QDesktopWidget()
            else:
                if 0:
                    print('get QDesktopWidget()')
                _qd = QtGui.QDesktopWidget()

            nmon = _qd.screenCount()
            wa = [_qd.screenGeometry(ii) for ii in range(nmon)]
            wa = [[w.x(), w.y(), w.width(), w.height()] for w in wa]

            if verbose:
                for ii, w in enumerate(wa):
                    print('monitor %d: %s' % (ii, str(w)))
        return wa
except:
    def monitorSizes(verbose=0):
        return [[0, 0, 1280, 720]]
    pass


def tilefigs(lst, geometry, ww=None, raisewindows=False, tofront=False, verbose=0):
    """ Tile figure windows on a specified area """
    mngr = plt.get_current_fig_manager()
    be = matplotlib.get_backend()
    if ww is None:
        ww = monitorSizes()[-1]

    w = ww[2] / geometry[0]
    h = ww[3] / geometry[1]

    # wm=plt.get_current_fig_manager()

    if verbose:
        print('tilefigs: ww %s, w %d h %d' % (str(ww), w, h))
    for ii, f in enumerate(lst):
        if not plt.fignum_exists(f):
            continue
        fig = plt.figure(f)
        iim = ii % np.prod(geometry)
        ix = iim % geometry[0]
        iy = np.floor(float(iim) / geometry[0])
        x = ww[0] + ix * w
        y = ww[1] + iy * h
        if verbose:
            print('ii %d: %d %d: f %d: %d %d %d %d' %
                  (ii, ix, iy, f, x, y, w, h))
            if verbose >= 2:
                print('  window %s' % mngr.get_window_title())
        if be == 'WXAgg':
            fig.canvas.manager.window.SetPosition((x, y))
            fig.canvas.manager.window.SetSize((w, h))
        if be == 'WX':
            fig.canvas.manager.window.SetPosition((x, y))
            fig.canvas.manager.window.SetSize((w, h))
        if be == 'agg':
            fig.canvas.manager.window.SetPosition((x, y))
            fig.canvas.manager.window.resize(w, h)
        if be == 'Qt4Agg' or be == 'QT4' or be == 'QT5Agg':
            # assume Qt canvas
            try:
                fig.canvas.manager.window.move(x, y)
                fig.canvas.manager.window.resize(w, h)
                fig.canvas.manager.window.setGeometry(x, y, w, h)
                # mngr.window.setGeometry(x,y,w,h)
            except Exception as e:
                print('problem with window manager: ', )
                print(be)
                print(e)
                pass
        if raisewindows:
            mngr.window.raise_()
        if tofront:
            plt.figure(f)


def plot2Dline(line, *args, **kwargs):
    """ Plot a 2D line in a matplotlib figure """
    if np.abs(line[1]) > .001:
        xx = plt.xlim()
        xx = np.array(xx)
        yy = (-line[2] - line[0] * xx) / line[1]
        plt.plot(xx, yy, *args, **kwargs)
    else:
        yy = np.array(plt.ylim())
        xx = (-line[2] - line[1] * yy) / line[0]
        plt.plot(xx, yy, *args, **kwargs)


#%% Make nice plots
# http://blog.olgabotvinnik.com/prettyplotlib/

def niceplot(ax, fig=None, despine=True, verbose=0, figurebg=True, tightlayout=True, legend=None, almost_black='#222222'):
    """ Create a good looking plot

    The code:
        - removes spines
        - makes legend and spines lighter
        - makes legend lighter

    """

    # Remove top and right axes lines ("spines")
    if verbose:
        print('niceplot: remove spines')

    spines_to_keep = ['bottom', 'left']
    if despine:
        spines_to_remove = ['top', 'right']
        for spine in spines_to_remove:
            ax.spines[spine].set_visible(False)
    else:
        spines_to_keep += ['top', 'right']
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    if verbose:
        print('niceplot: reduce spine intensity')

    for spine in spines_to_keep:
        ax.spines[spine].set_linewidth(0.5)
        ax.spines[spine].set_color(almost_black)

    ax.tick_params(axis='both', direction='out')

    if not legend is None:
        if verbose:
            print('niceplot: adjust legend')

        # Remove the line around the legend box, and instead fill it with a light grey
        # Also only use one point for the scatterplot legend because the user will
        # get the idea after just one, they don't need three.
        light_grey = np.array([float(241) / float(255)] * 3)
        rect = legend.get_frame()
        rect.set_facecolor(light_grey)
        middle_grey = np.array([float(151) / float(255)] * 3)
        rect.set_edgecolor(middle_grey)
        # rect.set_linewidth(0.0)

        # Change the legend label colors to almost black, too
        texts = legend.texts
        for t in texts:
            t.set_color(almost_black)

        # ttx=legend.get_texts()
        #[v.set_color(almost_black) for v in ttx]

    # fig.tight_layout(pad=0.1)
    if not fig is None and tightlayout:
        fig.tight_layout(pad=1.0)

    if figurebg and fig is not None:
        fig.set_facecolor('w')

    plt.draw()
    plt.show()


def enlargelims(factor=1.05):
    """ Enlarge the limits of a plot
    Example:
      >>> enlargelims(1.1)
    """
    xl = plt.xlim()
    d = (factor - 1) * (xl[1] - xl[0]) / 2
    xl = (xl[0] - d, xl[1] + d)
    plt.xlim(xl)
    yl = plt.ylim()
    d = (factor - 1) * (yl[1] - yl[0]) / 2
    yl = (yl[0] - d, yl[1] + d)
    plt.ylim(yl)

#%%


def selectParetoArrays(allarrays, pp):
    """ Select arrays using a Pareto object """
    paretoarrays = oalib.arraylist_t()
    paretoidx = np.array(pp.allindices())
    ww = oalib.longVector(tuple(paretoidx.tolist()))
    oalib.selectArrays(allarrays, ww, paretoarrays)
    return paretoarrays


def joinArrayLists(ww):
    """ Concetenate a list of array lists into a single list """
    ll = oalib.arraylist_t()
    for w in ww:
        for al in w:
            ll.push_back(al)
    return ll


def createPareto(dds, verbose=1):
    """ Create Pareto object from dataset """
    pp = oapackage.ParetoDoubleLong()

    for ii in range(dds.shape[0]):
        v = (dds[ii, :]).astype(np.float64)
        pp.addvalue(v, ii)

    if verbose:
        pp.show(verbose)
    return pp


#%% Utils


def static_var(varname, value):
    """ Helper function to create a static variable """
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate


@static_var("time", 0)
def tprint(string, dt=1, output=False):
    """ Print progress of a loop every dt seconds """
    if (time.time() - tprint.time) > dt:
        print(string)
        tprint.time = time.time()
        if output:
            return True
        else:
            return
    else:
        if output:
            return False
        else:
            return


def timeString(tt=None):
    """ Return a string with the current time or specified time """
    if tt is None:
        tt = gmtime()
    ts = strftime("%Y-%m-%d %H:%M:%S", tt)
    return ts


def findfilesR(p, patt):
    """ Get a list of files (recursive) """
    lst = []
    for root, dirs, files in os.walk(p, topdown=False):
        lst += [os.path.join(root, f) for f in files]
        # for name in files:
        #    lst += [name]
    rr = re.compile(patt)
    lst = [l for l in lst if re.match(rr, l)]
    return lst


def findfiles(p, patt=None):
    """ Get a list of files """
    lst = os.listdir(p)
    if not patt is None:
        rr = re.compile(patt)
        lst = [l for l in lst if re.match(rr, l)]
    return lst


def finddirectories(p, patt=None):
    """ Get a list of directories """
    lst = os.listdir(p)
    if not patt is None:
        rr = re.compile(patt)
        lst = [l for l in lst if re.match(rr, l)]
    lst = [l for l in lst if os.path.isdir(os.path.join(p, l))]
    return lst


def oainfo(afile, verbose=1):
    """ Print information about a file containing arrays """
    af = oapackage.arrayfile_t(afile, verbose)
    print(af.showstr())
    af.closefile()


def oaIsBinary(afile):
    """ Return true if array file is in binary format """
    af = oapackage.arrayfile_t(afile)
    ret = af.isbinary()
    af.closefile()
    return ret


def fac(n):
    """ Return n! (factorial) """
    if n == 1 or n == 0:
        return 1
    else:
        return n * fac(n - 1)


def choose(n, k):
    """ Return n choose k """
    ntok = 1
    for t in range(min(k, n - k)):
        ntok = ntok * (n - t) // (t + 1)
    return ntok
#    return fac(n)/(fac(n-k)*fac(k))


def array2latex(X, header=1, hlines=[], floatfmt='%g', comment=None, hlinespace=None, mode='tabular', tabchar='c'):
    """ Convert numpy array to Latex tabular """
    ss = ''
    if comment is not None:
        ss += '%% %s\n' % str(comment)
    if header:
        if mode == 'tabular':
            if len(tabchar) == 1:
                cc = tabchar * X.shape[1]
            else:
                cc = tabchar + tabchar[-1] * (X.shape[1] - len(tabchar))
            ss += '\\begin{tabular}{%s}' % cc + chr(10)
        elif mode == 'psmallmatrix':
            ss += '\\begin{psmallmatrix}' + chr(10)
        else:
            ss += '\\begin{pmatrix}' + chr(10)
        # ss += '\hline' + chr(10)
    for ii in range(X.shape[0]):
        r = X[ii, :]
#        for jj in range(X.shape[1]):
#            v=r[jj]
        if isinstance(r[0], str):
            ss += ' & '.join(['%s' % x for x in r])
        else:
            ss += ' & '.join([floatfmt % x for x in r])
        if ii < (X.shape[0]) - 1 or not header:
            ss += '  \\\\' + chr(10)
        else:
            ss += '  ' + chr(10)
        if ii in hlines:
            ss += '\hline' + chr(10)
            if hlinespace != None:
                ss += '\\rule[+%.2fex]{0pt}{0pt}' % hlinespace
    if header:
        if mode == 'tabular':
            ss += '\\end{tabular}'
        elif mode == 'psmallmatrix':
            ss += '\\end{psmallmatrix}' + chr(10)
        else:
            ss += '\\end{pmatrix}' + chr(10)
    return ss


def array2latex_old(ltable, htable=None):
    """ Write a numpy array to latex format """
    print('please use array2latex from researchOA instead')
    ss = ''
    ccc = 'c' * ltable.shape[1]
    ss += '\\begin{tabular}{%s}\n' % ccc
    if not htable is None:
        ss += " \\\\\n".join(
            [" & ".join(['\\textbf{%s}' % val for val in htable])])
        ss += ' \\\\ \n'
        ss += '\\hline \n'
    for ii in range(0, ltable.shape[0]):
        line = ltable[ii, :]
        ss += " & ".join([str(val) for val in line])
        ss += ' \\\\ \n'
    ss += '\\end{tabular}\n'
    return ss


def array2html(X, header=1, tablestyle='border-collapse: collapse;', trclass='', tdstyle='', trstyle='', thstyle=''):
    """ Convert Numpy array to HTML table

    Arguments
    ---------
        X : numpy array
            array to be converted
        header : integer
            use header or not
    Returns
    -------
        page : markup html object
            generated table in HTML

    """
    page = markup.page()
    page.add('<!-- Created by array2html -->\n')
    page.table(style=tablestyle)
    offset = 0
    nc = X.shape[1]
    nr = X.shape[0]

    if isinstance(trstyle, str):
        trstyle = [trstyle] * nr
    if isinstance(trclass, str):
        trclass = [trclass] * nr

    ri = 0
    if header:
        page.tr(style='font-weight: bold; border-bottom: solid 1px black;' +
                trstyle[ri], class_=trclass[ri])
        ri = ri + 1
        for ii in range(nc):
            if isinstance(X[offset, ii], tuple):
                print('array2html: tuple instance')
                page.th(X[offset, ii][0], style=thstyle + X[offset, ii][1])
            else:
                page.th(X[offset, ii], style=thstyle)
        page.tr.close()
        offset = offset + 1

    nr = X.shape[0] - offset
    for r in range(nr):
        page.tr(style=trstyle[ri], _class=trclass[ri])
        for ii in range(nc):
            if isinstance(X[offset, ii], tuple):
                page.td(X[offset, ii][0], style=tdstyle + X[offset, ii][1])
            else:
                page.td(X[offset, ii], style=tdstyle)

        page.tr.close()
        offset = offset + 1
        ri = ri + 1
    page.table.close()
    return page

import subprocess


def runcommand(cmd, dryrun=0, idstr=None, verbose=1, logfile=None, shell=True):
    """ Run specified command in external environment 

    Returns:
        r (int): return value of the shell command
    """
    if not idstr is None:
        cmd = 'echo "idstr: %s";\n' % idstr + cmd
    if verbose >= 2:
        print('cmd: %s' % cmd)
    r = 0
    if not dryrun:

        process = subprocess.Popen(
            cmd, bufsize=1, stdout=subprocess.PIPE, shell=shell)
        for jj in range(10000000):
            r = process.poll()
            #print('poll done... %s' % r)
            line = process.stdout.readline()
            # print('poll...')
            if verbose >= 2:
                print('runcommand: jj %d' % jj)
            if verbose >= 3:
                print('runcommand: jj %d: "%s"' % (jj, line))
            if len(line) == 0:
                if verbose >= 2:
                    print('runcommand: len(line) %d' % len(line))

                break

            if r is not None:
                break
            if jj > 20000000:
                print('error: early abort of runcommand')
                break
            line = line.decode(encoding='UTF-8')
            sys.stdout.write(str(line))
            if jj % 2 == 0:
                sys.stdout.flush()
            if verbose >= 2:
                print('end of loop...')
        #print('exit loop...')
        r = process.poll()
        # r = os.system(cmd) # old method
        if (not r == 0):
            print('runcommand: cmd returned error! r=%d' % str(r) )
            print(cmd)
            return r
    else:
        if verbose >= 2 or (verbose and logfile is None):
            print('### dryrun\n%s\n###\n' % cmd)
    if logfile is not None:
        fid = open(logfile, 'a')
        fid.write('\n###\n')
        fid.write(cmd)
        fid.close()
    else:
        pass
        #print('no logfile')
        #raise Exception('no logfile')

    # all good
    return r


def getArrayFile(afile):
    """ Return pointer to array file
    Automatically check for compressed array files
    """
    afilegz = afile = afile + '.gz'
    if not os.path.exists(afile) and os.path.exists(afilegz):
        afile = afilegz
    else:
        afile = afile
    return afile


def checkOAfile(afile, verbose=0):
    """ Return pointer to array file
    Automatically check for compressed array files
    """
    af = oapackage.arrayfile_t(afile, verbose)
    nm = af.filename
    if af.isopen():
        return nm
    else:
        return None


def checkArrayFile(afile, cache=1):
    """ Check whether a file or list of files exists
        If the file does not exist, check whether a compressed file does exist
        cache: 0 (no), 1 (check), -1 (always)
    """
    if cache == -1:
        return True
    if cache == 0:
        return False

    if os.path.exists(afile):
        return True
    else:
        if afile.endswith('.oa'):
            if os.path.exists(afile + '.gz'):
                return True
            else:
                return False
        else:
            return False

try:
    basestring  # attempt to evaluate basestring

    def isstr(s):
        return isinstance(s, basestring)
except NameError:
    # probably Python 3.x
    def isstr(s):
        return isinstance(s, str)


def checkFiles(lst, cache=1, verbose=0):
    """ Check whether a file or list of files exists
        cache: 0 (no), 1 (check), -1 (always)

        Returns False if one or more of the files do not exist
        Returns True if all files exist
    """
    if cache == -1:
        return True
    if cache == 0:
        return False

    if isstr(lst):
        lst = [lst]
    c = True
    for f in lst:
        if not os.path.exists(f):
            if verbose:
                print('checkFiles: file %s does not exist' % f)
            c = False
            break
    return c

#%%


def checkFilesOA(lst, cache=1, verbose=0):
    """ Check whether a file or list of files exists
        
        Args:
            lst (list)
            cache (int): 0 (always return False), 1 (check), -1 (always return True)
        For array files also the .gz extension is checked

        Returns False if one or more of the files do not exist
        Returns True if all files exist
    """
    if verbose>=2:
        print('checkFilesOA: cache %s'  % (cache, ))
    if cache == -1:
        return True
    if cache == 0:
        return False

    if isstr(lst):
        lst = [lst]
    c = True
    for f in lst:
        r = checkArrayFile(f, cache=True)
        if not r:
            if verbose:
                print('checkFiles: file %s does not exist' % f)
            c = False
            break
    return c

#%%

# import gc


def randomizearrayfile(afile, afileout, verbose=1):
    """ Randomize a file with arrays """
    lst = oapackage.readarrayfile(afile)
    rlst = oapackage.oalib.arraylist_t()
    for ii, al in enumerate(lst):
        adata = oalib.arraylink2arraydata(al)
        trans = oalib.array_transformation_t(adata)
        if verbose >= 2:
            print('randomizing array %d\n' % ii)
        trans.randomize()
        alr = trans.apply(al)
        rlst.push_back(alr)
    oapackage.writearrayfile(afileout, rlst)


def nArrayFile(afile, verbose=1):
    """ Return number of arrays in file """
    af = oalib.arrayfile_t(afile, 0)
    n = af.narrays
    af.closefile()
    return n


def selectArrays(infile, outfile, idx, afmode=oalib.ATEXT, verbose=1, cache=1):
    """ Select arrays in a file by indices

    Input:
        - infile (string)
        - outfile (string)
        - inx (list of indices)
    Output:
        - None

    """
    if not checkFiles(outfile, cache=cache):
        gidxint = oalib.intVector([int(x) for x in idx])
        # print('  xx select: %s' % str(gidxint))
        # print('  xx select: %s' % str([v for v in gidxint]))
        sols = oalib.arraylist_t()
        oalib.selectArrays(infile, gidxint, sols, 0)
        af = oalib.arrayfile_t(infile, 1)
        af.nrows
        r = oalib.writearrayfile(outfile, sols, afmode, af.nrows, af.ncols)
        if verbose >= 2:
            print('selectArrays: write array file %s in mode %d' %
                  (outfile, afmode))
    else:
        if verbose >= 2:
            print('output file %s already exists' % outfile)


def floatformat(number, mind=2, maxd=4):
    """ Format a floating point number into a string """
    rounded = int(number * (10**mind)) / float(10**mind)
    if number == rounded:
        return ('%%.0%df' % mind) % number
    else:
        return ('%%.0%dg' % maxd) % number


def safemin(data, default=0):
    """ Return mininum of array with default value for empty array """
    if data.size == 0:
        m = default
    else:
        m = data.min()
    return m


def mkdirc(xdir):
    """ Create directory """
    if not os.path.exists(xdir):
        os.mkdir(xdir)
    return xdir


def parseProcessingTime(logfile, verbose=0):
    """ Parse a log file to calculate the processing time """

    try:
        import dateutil.parser
    except:
        warnings.warn('oahelper: could not load dateutil package...')
        pass

    fileinput.close()
    tstart = None
    tend = None
    dtr = None
    try:
        for line in fileinput.input([logfile]):
            if line.startswith('#time'):
                if verbose >= 1:
                    # print(line)
                    print('parseProcessingTime: line: %s' % line, end="")
                    # print('xy %s' % line[10:])
            if line.startswith('#time start:'):
                tstart = dateutil.parser.parse(line[13:])
                if verbose >= 2:
                    print('parseProcessingTime: tstart: %s' % tstart)
            # else:
            #    print('invalid line? %s' % line)
            if line.startswith('#time end:'):
                tend = dateutil.parser.parse(line[10:])
                if verbose >= 2:
                    print('parseProcessingTime: tend: %s' % tend)
            if line.startswith('#time total:'):
                dtr = (line[13:])
                dtr = float(dtr[:-5])
                if verbose >= 2:
                    print('parseProcessingTime: total: %s' % dtr)
            # else:
            #    print('invalid line? %s' % line)
        if tstart is not None and tend is not None:
            dt = tend - tstart
            dtt = dt.total_seconds()
        else:
            dtt = -1
    except:
        if verbose:
            print('error processing log %s' % logfile)
            traceback.print_exc(file=sys.stdout)
        dtt = -1
    if not dtr is None:
        if abs(dtr - dtt) > 10:
            print(
                'parseProcessingTime: warning difference in reported times %.1f dtr %.1f [s]' % (dtt, dtr))
    return dtt


def safemax(data, default=0):
    """ Return maximum of array with default value for empty array """

    if isinstance(data, list):
        if len(data) == 0:
            m = default
        else:
            m = max(data)
        return m
    if data.size == 0:
        m = default
    else:
        m = data.max()
    return m


def series2htmlstr(ad, html=1, case=0):
    """ Convert arraydata_t to html formatted string """
#    s=[ad.s[ii] for ii in range(0, ad.ncols)]
    s = list(ad.getS())
    p = -1
    n = 0
    aa = []
    bb = []
    while n < len(s):
        if s[n] != p:
            p = s[n]
            aa.append(p)
            bb.append(1)
        else:
            bb[-1] += 1
        n = n + 1
    if case == 0:
        if bb[-1] > 1:
            bb[-1] = 'a'
    hstr = 'OA(%d; %d; ' % (ad.N, ad.strength)
    for ii in range(0, len(aa)):
        if html:
            hstr += '%d<sup>%s</sup>' % (aa[ii], str(bb[ii]))
        else:
            hstr += '%d^%s' % (aa[ii], str(bb[ii]))
    hstr += ')'
    return hstr


def gwlp2str(gmadata, t=None, sformat=None, jstr=','):
    """ Convert GWLP value to string format """
    if gmadata is None:
        return '-'
    if isinstance(gmadata, tuple):
        # do nothing
        gmadata
    else:
        if isinstance(gmadata, list):
            # do nothing
            gmadata
        else:
            gmadata[gmadata < 0] = 0
        if not(np.abs(gmadata[0] - 1) < 1e-12 and np.abs(gmadata[1]) < 1e-12):
            print('warning: data are not good GWPL data!!!!')
            return ''
            # pdb.set_trace()
    bgma = np.around(gmadata, decimals=12)
    if not t is None:
        bgma = bgma[(t + 1):]
    if sformat is None:
        gstr = jstr.join([floatformat(v, mind=2, maxd=4) for v in bgma])
    else:
        gstr = jstr.join([sformat % v for v in bgma])
    return gstr


def selectJ(sols0, jj=5, jresults=None, verbose=1):
    """ Select only arrays with J-characteristics non-zero """
    if jresults is None:
        jresults = oalib.analyseArrays(sols0, verbose, jj)

    solseo = oalib.arraylist_t()
    v = []
    for ii, js in enumerate(jresults):
        # js.show()
        #    if not np.array(js.vals)[0]==0:
        #        solseo.append(sols0[ii])
        v.append(js.vals[0])

    si = [i for (i, j) in sorted(enumerate(v), key=operator.itemgetter(1))]
    for jj in si:
        if v[jj] > 0:
            solseo.append(sols0[jj])
    if verbose:
        print('selectJ: kept %d/%d solutions' % (solseo.size(), sols0.size()))
    return solseo


def extendSingleArray(A, adata, t=3, verbose=1):
    """ Extend a single array """
    oaoptions = oalib.OAextend()
    sols0 = oalib.arraylist_t()
    solsx = oalib.arraylist_t()
    sols0.push_back(A)
    #N = A.n_rows
    k = A.n_columns
    n = oalib.extend_arraylist(sols0, adata, oaoptions, k, solsx)
    sys.stdout.flush()
    return solsx


def runExtend(N, k, t=3, l=2, verbose=1, initsols=None, nums=[], algorithm=None):
    """ Run extension algorithm and return arrays

    Arguments
    ---------
    N : integer
        number of rows
    k: integer
        number of columns
    t: integer
        strength of the arrays

    >>> r = runExtend(16, 5, 3, verbose=0)    
    """
    if verbose:
        print('runExtend: N=%d, k=%d, t=%d' % (N, k, t))
    if isinstance(l, list):  # types.ListType):
        ll = l
    else:
        ll = [l]
    ll = ll + [ll[-1]] * (k - len(ll))
    s = oalib.intVector(ll)
    adata = oalib.arraydata_t(s, N, t, k)
    al = oalib.array_link(adata.N, adata.strength, 1)
    al.create_root(adata)
    if initsols is None:
        sols0 = oalib.arraylist_t()
        sols0.append(al)
        tstart = t
    else:
        sols0 = initsols
        tstart = sols0[0].n_columns

    oaoptions = oalib.OAextend()
    if algorithm is None:
        oaoptions.setAlgorithmAuto(adata)
    else:
        oaoptions.setAlgorithm(algorithm, adata)
    solsx = sols0
    for ii in range(tstart, k):
        solsx = oalib.arraylist_t()
        oalib.extend_arraylist(sols0, adata, oaoptions, ii, solsx)
        if verbose >= 2:
            print(' ii %d: %d' % (ii, solsx.size()))
        sols0 = solsx
        nums.append(solsx.size())
        sys.stdout.flush()
    return solsx


def compressOAfile(afile, decompress=False, verbose=1):
    """ Compress an OA array file """
    af = oalib.arrayfile_t(afile, 0)
    if verbose >= 2:
        print('file %s: binary %s' % (afile, af.isbinary()))
    if not sys.platform == 'linux2':
        if verbose:
            print('compressOAfile: not compressing file %s (platform not supported)' %
                  afile)
        return False
    if af.isbinary() and not af.iscompressed and sys.platform == 'linux2':
        if verbose:
            print('compressOAfile: compressing file %s' % afile)
        cmd = 'gzip -f %s' % afile
        r = os.system(cmd)
        if not r == 0:
            print('compressOAfile: error compressing file %s' % afile)
        return True
    else:
        if not af.isbinary():
            if verbose:
                print('compressOAfile: not compressing file %s (file is in text mode)' %
                      afile)
        else:
            if verbose:
                print('compressOAfile: not compressing file %s (file %s is compressed)' %
                      (afile, af.filename))
        return False

if 0:
    def pointer2np(p, sz):
        """ Convert array pointer to numpy array """
        if isinstance(sz, tuple):
            1
        else:
            sz = (sz,)
        ip = oalib.intArray.frompointer(p)
        A = np.zeros(sz)
        for ii in range(0, np.prod(sz)):
            A[ii] = ip[ii]
        A = A.reshape(sz)
        return A

if 0:
    import ctypes

    def arraylink2array(al):
        """ Convert arraylink to numpy array """
        print('use araylink.getarray')
        w = al.array
        nrows = al.n_rows
        ncols = al.n_columns
        func = ctypes.pythonapi.PyBuffer_FromMemory
        func.restype = ctypes.py_object
        buffer = func(w.__long__(), nrows * ncols * 2 * 2)
        xx = np.frombuffer(buffer, np.int32, count=nrows * ncols)
        xx = xx.reshape((ncols, nrows)).transpose()
        return xx


def getposjvals(A, t, verbose=1):
    N = A.shape[0]
    k = A.shape[1]
    ncols = t + 1
    jstep = 2**(t + 1)
    njvals = int(1 + (float(N) / jstep))
    jvals = [N - jstep * l for l in range(0, njvals)]
    return jvals


def arraystats(A, verbose=1):
    """ Return statistics of an array """
    Af = A.transpose().flatten()
    al = oalib.array_link(A.shape[0], A.shape[1], 0)
    alist = oalib.arraylist_t(1)
    alist[0] = al
    # al.array
    ia = oalib.intArray.frompointer(al.array)
    for x in range(0, A.size):
        ia[x] = int(Af[x])
#    al.showarray()
    jresults = oalib.analyseArrays(alist, 0)
    js = jresults[0]
    vals = pointer2np(js.vals, js.nc)
    jvals = getposjvals(A, 3, verbose=0)

    jv = np.zeros(len(jvals))
    c = Counter(abs(vals))
    for ii, xx in enumerate(jvals):
        if c.has_key(xx):
            jv[ii] = c.get(xx)
    if verbose:
        print('Possible j-values: %s' % jvals, end="")
        print('     values: %s' % jv.astype(int))

    N = A.shape[0]
    Ak = (1 / N ^ 2) * sum(vals**2)
    print('Ak: %s' % Ak)
    return Ak


def argsort(seq):
    """ Stable argsort """
    # http://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python/3382369#3382369
    return sorted(range(len(seq)), key=seq.__getitem__)


def jseq(xx, comb):
    pp = functools.reduce(lambda y, i: xx[:, i] + y, comb, 0)
    jseq = 1 - 2 * np.mod(pp, 2)
    return jseq


def sortrows(x):
    """ Sort rows of an array, return indices"""
    if len(x.shape) == 1:
        nn = 1
        sind = np.argsort(x)  # or np.argsort(x, order=('x', 'y'))
        return sind
    else:
        nn = x.shape[1]
    dt = [('x%d' % xx, 'f8') for xx in range(0, nn)]
    if x.size == 0:
        # hack
        sind = np.array([])
        return sind
    v = [tuple(x[kk, :]) for kk in range(0, x.shape[0])]
    w = np.array(v, dtype=dt)
    sind = np.argsort(w)  # or np.argsort(x, order=('x', 'y'))
    return sind


def sortcols(X):
    """ Sort columns of an array, return indices"""
    sind = sortrows(X.transpose())
    return sind


def showtriangles(jresults, showindex=1):
    """ Show triangle of j-values """
    if isinstance(jresults, oalib.jstruct_t):
        showindex = 0
        jresults = (jresults,)

    js = jresults[0]
    i = 0
    idx = [i]
    for j in range(4, js.k + 1):
        nn = choose(j - 1, 3)
        i = i + nn
        idx.append(i)
    for jj, js in enumerate(jresults):
        vals = oalib.intArray.frompointer(js.vals)
        if showindex:
            print('i: %d' % jj)
        for kk in range(0, js.k - 3):
            xx = [vals[v] for v in range(idx[kk], idx[kk + 1])]
            s = ','.join(map(str, xx))
            print('%s' % s)
#%%

from oapackage import markup
import webbrowser
import tempfile


def testHtml(hh=None):
    """ Test a short snippet of HTML """
    if hh is None:
        return
    page = markup.page()
    page.init()
    page.body()
    page.add(hh)
    page.body.close()
    tmp, f = tempfile.mkstemp('.html')
    with open(f, 'wt') as fname:
        fname.write(str(page))
        fname.close()
        webbrowser.open(fname.name)


def designStandardError(al):
    """ Return standard errors for a design

    Arguments
    ---------
    al : array
            design

    Output
    ------
    m0, m1, m2 : arrays
            standard errors

    """

    X = np.matrix(al.getModelMatrix(2))
    k = al.n_columns

    m = 1 + k + k * (k - 1) / 2

    # scalefac=np.sqrt(al.n_rows)
    scalefac = 1
    M = (X.transpose() * X / scalefac).I

    mm = np.array(M.diagonal()).flatten()

    m1 = mm[1:(1 + k)].flatten()
    m1 = m1[np.argsort(m1)]
    m2 = mm[(1 + k):]
    m2 = m2[np.argsort(m2)]
    m0 = mm[0]
    return np.sqrt(m0), np.sqrt(m1), np.sqrt(m2)
    # return m0,m1, m2


#%%
def DefficiencyBound(D, k, k2):
    """ Calculate the D-efficiency bound of an array extension

    Input
    -----
    D : float
        D-efficiency of the design
    k, k2: integers
        numbers of columns 
    Output
    ------
    D2 : float
        bound on the D-efficiency of extensions of a design with k columns to k2 columns    

    """
    m = 1. + k + k * (k - 1) / 2
    m2 = 1. + k2 + k2 * (k2 - 1) / 2
    D2 = D**(m / m2)
    return D2

#%% Misc


def setWindowRectangle(x, y=None, w=None, h=None, mngr=None, be=None):
    """ Position the current Matplotlib figure at the specified position
    Usage: setWindowRectangle(x,y,w,h)
    """
    if y is None:
        y = x[1]
        w = x[2]
        h = x[3]
        x = x[0]
    if mngr is None:
        mngr = plt.get_current_fig_manager()
    be = matplotlib.get_backend()
    if be == 'WXAgg':
        mngr.canvas.manager.window.SetPosition((x, y))
        mngr.canvas.manager.window.SetSize((w, h))
    elif be == 'agg':
            # mngr.canvas.manager.window.setGeometry(x,y,w,h)
        mngr.canvas.manager.window.SetPosition((x, y))
        mngr.canvas.manager.window.resize(w, h)
    elif be == 'module://IPython.kernel.zmq.pylab.backend_inline':
        pass
        # mngr.canvas.manager.window.setGeometry(x,y,w,h)
        #mngr.canvas.manager.window.SetPosition((x, y))
        #mngr.canvas.manager.window.resize(w, h)
    else:
        # assume Qt canvas
        mngr.canvas.manager.window.move(x, y)
        mngr.canvas.manager.window.resize(w, h)
        mngr.canvas.manager.window.setGeometry(x, y, w, h)
        # mngr.window.setGeometry(x,y,w,h)


def makearraylink(al):
    """ Convert array to array_link object """
    if isinstance(al, np.ndarray):
        tmp = oalib.array_link()
        tmp.setarray(al)
        al = tmp
    return al


def formatC(al, wrap=True):
    """ Format array the inclusion in C code """
    l = np.array(al).T.flatten().tolist()
    s = ','.join(['%d' % x for x in l])
    if wrap:
        s = '\tarray_link al ( %d,%d, 0 );\n\tint tmp[] = {' % (al.n_rows, al.n_columns) + s + '};'
    return s

#%%


def create_pareto_element(values, pareto=None):
    """ Create a vector of mvalue_t elements 
    Args:
        vv (list): list with tuples or arrays
    """
    if pareto is None:
        pareto = oalib.ParetoDoubleLong()
    if isinstance(pareto, oalib.ParetoMultiLongLong):
        vector_pareto = oalib.mvalueVector()
        for v in values:
            if isinstance(v, (int)):
                # convert to list type
                v = [v]
            if not isinstance(v, (list, type)):
                raise Exception('creating Pareto element for Pareto object of type %s and input of type %s not supported' % (type(pareto), type(v)))
            vec = oalib.mvalue_t_long(list(v))
            vector_pareto.push_back(vec)
    elif isinstance(pareto, oalib.ParetoMultiDoubleLong):
        vector_pareto = oalib.vector_mvalue_t_double()  # FIXME: naming of GWLPvalueVector
        for v in values:
            if isinstance(v, (int, float)):
                # convert to list type
                v = [float(v)]
            if not isinstance(v, (list, type)):
                raise Exception('creating Pareto element for Pareto object of type %s and input of type %s not supported' % (type(pareto), type(v)))

            vec = oalib.mvalue_t_double(list(v))
            vector_pareto.push_back(vec)
    elif isinstance(pareto, oalib.ParetoDoubleLong):
        if not isinstance(values, (list, tuple, np.ndarray)):
            raise('')
        vector_pareto = values
    else:
        raise Exception('creating Pareto element for Pareto object of type %s and input of type %s not supported' % (type(pareto), type(v)))
    return vector_pareto



# Orthogonal Array package
# pieter.eendebak@gmail.com

__all__ = ['oahelper']
__description__ = "Orthogonal Array package"
__uri__ = "http://www.pietereendebak.nl/oapackage/index.html"
__doc__ = __description__ + " <" + __uri__ + ">"

import oalib
oalib.setloglevel(oalib.SYSTEM)
oalib.log_print(-oalib.SYSTEM, '')
from oalib import *
# import .oahelper
from . oahelper import *
# import .Doptim
from . Doptim import *
from . import scanf

import numpy as np

__version__ = oalib.version()

#%%


def autodoctest():
    """ Test the module using autodoc
    Example:
      >>> import oapackage
      >>> arrayclass=oapackage.arraydata_t(2, 40, 0, 7)
      >>> print(arrayclass)
      arrayclass: N 40, k 7, strength 0, s {2,2,2,2,2,2,2}, order 0

    """
    return


def test_numpy_interface(verbose=0):
    A = np.eye(3, 4).astype(int)
    A[0, :] = [10, 20, 30, 50]
    A[2, :] = [-3, -4, -5, -6]

    if verbose:
        print('makearraylink')
    al = makearraylink(A)
    if verbose:
        al.showarray()

    if verbose:
        print('direct')
    al = array_link(A)
    if verbose:
        al.showarray()
    Ax = np.array(al)
    if verbose:
        print(A)
        print(Ax)

    if 0:
        # not possible right now...
        if verbose:
            print('direct float')
        A = np.eye(3).astype(float)
        al = array_link(A)
        if verbose:
            al.showarray()


def test_nauty(verbose=0):
    if verbose:
        print('test_nauty: test reduction to normal form')
    al = oalib.exampleArray(1, verbose)
    alr = al.randomperm()
    tr = oalib.reduceOAnauty(alr)
    alx = tr.apply(alr)
    assert(alx == al)


def unittest(verbose=1):
    """ Perform some unit testing, return True if succesfull """
    if verbose:
        print('oapackage: unittest: oalib version %s' % oalib.version())
    al = oalib.array_link()
    ii = 0
    al = oalib.exampleArray(ii, 0)

    arrayclass = oalib.arraydata_t(2, 16, 2, 6)  # fine
    alist = [arrayclass.create_root()]
    r = oalib.extend_arraylist(alist, arrayclass)
    if len(r) != 3:
        raise Exception('extension generation for %s failed' % (arrayclass, ))

    if not isinstance(al.getarray(), np.ndarray):
        print(
            'oapackage: unittest: error: array interface not working properly')
    else:
        if not al[2, 0] == al.getarray()[2, 0]:
            print(
                'oapackage: unittest: error: array interface not working properly')

    arrayclass = oalib.arraylink2arraydata(al)
#    print(arrayclass)

    if verbose >= 2:
        print('unittest: calculate efficiencies')
    Deff = al.Defficiency()
    aa = oalib.Aefficiencies(al)
    if verbose >= 2:
        print('## oapackage test: example array %d: Deff %.3f' % (ii, Deff))

    # DOP reduction
    if verbose >= 2:
        print('unittest: test delete-one-factor GWLP reduction')
    al = oalib.exampleArray(5, 1)
    al2 = al.randomperm()

    alr = al.reduceDOP()
    al2r = al2.reduceDOP()
    if not alr == al2r:
        print('error: DOP reduced arrays unequal!: %d' % (alr == al2r))
        print('alr')
        alr.showarraycompact()
        print('al2r')
        al2r.showarraycompact()
        return False

    at = oalib.reductionDOP(al)
    check = at.apply(al) == al.reduceDOP()
    if not check:
        print('error: DOP reduction transformation is invalid')

    # test graphtools
    if verbose >= 2:
        print('unittest: test graphtools')
    # import . graphtools
    from . graphtools import oa2graph
    arrayclass = oalib.arraylink2arraydata(al)
    _ = oa2graph(al, arrayclass)

    test_numpy_interface()
    test_nauty()

    return True

#%%
if __name__ == "__main__":
    """ Dummy main for oapackage """
    import doctest
    doctest.testmod()
